/******************************************************************************
 *  LMBFeedforward.h - Classes para rede neural feedforward                   *                   *
 *  Copyright(c) 1998 by LMBarros - Todos os direitos e esquerdos reservados  *
 *  Utiliza a classe TFileStream da VCL do C++ Builder...                     *
 ******************************************************************************/

#ifndef LMBFeedforward_h
#define LMBFeedforward_h

#include <math.h>

// ============================================================================
// = Funes Independentes                                                    =
// ============================================================================

// ----------------------------------------------------------------------------
// - Funo RandomWeight - Retorna um peso aleatrio                          -
// ----------------------------------------------------------------------------
inline float RandomWeight()
{
  float Num;
  Num = rand() % 1000;
  return 2 * (Num / 1000) - 1; // entre -1 e +1
}

// ----------------------------------------------------------------------------
// - Funo DotProd - Faz o produto interno de dois vetores de tamanho sz     -
// ----------------------------------------------------------------------------
float DotProd (int sz, float *vet1, float *vet2)
{

  int k, m, p = 0;
  float sum = 0;
  k = sz / 4; // Nmero de grupos de 4
  m = sz % 4; // Resto

  while (k--) // Faz os grupos de 4
  {
    sum += vet1[p] * vet2[p++];  // Fazendo desta forma, de 4 em 4, o nmero
    sum += vet1[p] * vet2[p++];  // de loops  reduzido. Isto deve acelerar
    sum += vet1[p] * vet2[p++];  // bastante o processo, especialmente se o
    sum += vet1[p] * vet2[p++];  // processador utilizado usar pipelining.
  }

  while (m--) // Faz o resto
    sum += vet1[p] * vet2[p++];

  return sum;
}

// ----------------------------------------------------------------------------
// - Funes para ativao - Usando a funo sigmoidal  1/(1+exp(-x))         -
// ----------------------------------------------------------------------------
/* Para obter um cdigo mais rpido, no se utiliza a funo diretamente, mas
   um esquema com tabelas com valores pr-calculados. A funo InitActFunc deve
   ser a primeira a ser chamada; ela inicializa as tais tabelas. */

#define F_TABLE_LENGTH 100
#define F_TABLE_MAX 10.0
static float f_factor, f_f[F_TABLE_LENGTH], f_d[F_TABLE_LENGTH];

// - InitActFunc - Inicializao das tabelas ----------------------------------
void InitActFunc()
{
  f_factor = (float)(F_TABLE_LENGTH - 1) / (float)F_TABLE_MAX;
  for (int c = 0 ; c < F_TABLE_LENGTH ; c++)
  {
    f_f[c] = 1.0 / (1.0 + exp (-((float)c) / f_factor));
    if (c) f_d[c-1] = f_f[c] - f_f[c-1];
  }
}

// - ActFunc - Faz a funo sigmoidal -----------------------------------------
float ActFunc (float x)
{
  int i;
  float xd; // distancia no eixo x

  if (x >= 0)
  {
    xd = x * f_factor;
    i = (int)xd;
    if (i >= (F_TABLE_LENGTH - 1)) return f_f[F_TABLE_LENGTH - 1];
    return f_f[i] + f_d[i] * (xd - i);
  }
  else
  {
    xd = -x * f_factor; // x  negativo, logo, xd  positivo (f_factor  positivo)
    i = (int)xd;
    if (i >= (F_TABLE_LENGTH - 1)) return 1.0 - f_f[F_TABLE_LENGTH - 1];
    return 1.0 - f_f[i] + f_d[i] * (xd - i);
  }
}

// - ActDeriv - Faz a derivada da funo de ativao --------------------------
float ActDeriv(float f) // O parmetro  f(x), no simplesmente x
{
  return f * (1.0 - f);
}

// ============================================================================
// = LMBFeedforwardInputLayer                                                 =
// ============================================================================

// ----------------------------------------------------------------------------
// - Interface                                                                -
// ----------------------------------------------------------------------------

class LMBFeedforwardInputLayer
{
 public:
  LMBFeedforwardInputLayer(int sz);
  ~LMBFeedforwardInputLayer();
  float *Outputs;
  int NumNrns;
};

// ----------------------------------------------------------------------------
// - Implementao                                                            -
// ----------------------------------------------------------------------------

// - Construtor ---------------------------------------------------------------
LMBFeedforwardInputLayer::LMBFeedforwardInputLayer(int sz)
{
  Outputs = new float[sz];
  NumNrns = sz;
}

// - Destrutor ----------------------------------------------------------------
LMBFeedforwardInputLayer::~LMBFeedforwardInputLayer()
{
  delete[] Outputs;
}

// ============================================================================
// = LMBFeedforwardLayer - Output e Hidden com outro nome...                  =
// ============================================================================

// ----------------------------------------------------------------------------
// - Interface                                                                -
// ----------------------------------------------------------------------------

class LMBFeedforwardLayer
{
 public:
  LMBFeedforwardLayer(int szThis, int szPrev);
  ~LMBFeedforwardLayer();
  void ForwardPropagate();
  void RandomizeWeights();
  float *Inputs;
  float *Outputs;
  float *Weights;
  int NumNrns;
  int NumNrnsPrev;
};

// ----------------------------------------------------------------------------
// - Implementao                                                            -
// ----------------------------------------------------------------------------

// - Construtor ---------------------------------------------------------------
LMBFeedforwardLayer::LMBFeedforwardLayer(int szThis, int szPrev)
{
  Outputs = new float[szThis];
  Weights = new float[szThis * szPrev];
  NumNrns = szThis;
  NumNrnsPrev = szPrev;
}

// - Destrutor ----------------------------------------------------------------
LMBFeedforwardLayer::~LMBFeedforwardLayer()
{
  delete[] Outputs;
  delete[] Weights;
}

// - ForwardPropagate ---------------------------------------------------------
void LMBFeedforwardLayer::ForwardPropagate()
{
  float *vcWeights;

  vcWeights = new float[NumNrnsPrev];

  for (int eON = 0 ; eON < NumNrns ; eON++) // Para cada neurnio da camada
  {
    // Prepara o vetor de pesos
    for (int eIN = 0 ; eIN < NumNrnsPrev ; eIN++)
      vcWeights[eIN] = Weights[NumNrnsPrev * eON + eIN];

    Outputs[eON] = ActFunc(DotProd(NumNrnsPrev, Inputs, vcWeights));
  }

  // Limpa
  delete[] vcWeights;
}

// - RandomizeWeights ---------------------------------------------------------
void LMBFeedforwardLayer::RandomizeWeights()
{
  for (int c = 0 ; c < NumNrns * NumNrnsPrev ; c++)
    Weights[c] = RandomWeight();
}

// ============================================================================
// = LMBFeedforwardOutputLayer                                                =
// ============================================================================

typedef LMBFeedforwardLayer LMBFeedforwardOutputLayer;

// ============================================================================
// = LMBFeedforwardHiddenLayer                                                =
// ============================================================================

typedef LMBFeedforwardLayer LMBFeedforwardHiddenLayer;

// ============================================================================
// = LMBFeedforwardNetwork                                                    =
// ============================================================================

// ----------------------------------------------------------------------------
// - Interface                                                                -
// ----------------------------------------------------------------------------

class LMBFeedforwardNetwork
{
  friend class LMBackprop;
 public:
  LMBFeedforwardNetwork(int szInp, int szHid, int szOut);
  ~LMBFeedforwardNetwork();
  float *GetOutputs(float *Inputs);
  void RandomizeWeights();
  void WriteWeights (char *FileName);
  bool ReadWeights (char *FileName); // retorna true se OK
  LMBFeedforwardInputLayer *InputLayer;
  LMBFeedforwardHiddenLayer *HiddenLayer;
  LMBFeedforwardOutputLayer *OutputLayer;
};

// ----------------------------------------------------------------------------
// - Implementao                                                            -
// ----------------------------------------------------------------------------

// - Construtor ---------------------------------------------------------------
LMBFeedforwardNetwork::LMBFeedforwardNetwork(int szInp, int szHid, int szOut)
{
  // Cria as camadas:
  InputLayer = new LMBFeedforwardInputLayer(szInp);
  HiddenLayer = new LMBFeedforwardHiddenLayer(szHid, szInp);
  OutputLayer = new LMBFeedforwardOutputLayer(szOut, szHid);
  // Conecta as camadas:
  HiddenLayer->Inputs = InputLayer->Outputs;
  OutputLayer->Inputs = HiddenLayer->Outputs;
  // Inicializao da funo de ativao e nmeros aleatrios
  InitActFunc();
//  srand(GetTickCount());
}

// - Destrutor ----------------------------------------------------------------
LMBFeedforwardNetwork::~LMBFeedforwardNetwork()
{
  delete InputLayer;
  delete HiddenLayer;
  delete OutputLayer;
}

// - GetOutputs ---------------------------------------------------------------
float *LMBFeedforwardNetwork::GetOutputs(float *Inputs)
{
  for (int c = 0 ; c < InputLayer->NumNrns ; c++)
    InputLayer->Outputs[c] = Inputs[c];
  HiddenLayer->ForwardPropagate();
  OutputLayer->ForwardPropagate();
  return OutputLayer->Outputs;
}

// - RandomizeWeights ---------------------------------------------------------
void LMBFeedforwardNetwork::RandomizeWeights()
{
  HiddenLayer->RandomizeWeights();
  OutputLayer->RandomizeWeights();
}

// -- WriteWeights ------------------------------------------------------------
void LMBFeedforwardNetwork::WriteWeights (char* FileName)
{
  // Consulte Apndice 1 para ver o formato dos pesos gravados em disco
  TFileStream *File;
  unsigned short int *iBuf;
  float *fBuf;
  unsigned long int Pos = 0;

  File = new TFileStream (FileName, fmCreate);

  // Escreve o cabealho:
  iBuf = new unsigned short int [3];
  iBuf[0] = (unsigned short int)InputLayer->NumNrns;
  iBuf[1] = (unsigned short int)HiddenLayer->NumNrns;
  iBuf[2] = (unsigned short int)OutputLayer->NumNrns;

  File->Write((void*)iBuf, 6); // So 3 unsigned short int's -> 6 bytes

  // Escreve os pesos:
  fBuf = new float [ (iBuf[0] * iBuf[1]) + (iBuf[1] * iBuf[2]) ];

  for (int eHN = 0 ; eHN < HiddenLayer->NumNrns ; eHN++)
    for (int eIN = 0 ; eIN < InputLayer->NumNrns ; eIN++)
      fBuf[Pos++] = HiddenLayer->Weights[InputLayer->NumNrns * eHN + eIN];

  for (int eON = 0 ; eON < OutputLayer->NumNrns ; eON++)
    for (int eHN = 0 ; eHN < HiddenLayer->NumNrns ; eHN++)
      fBuf[Pos++] = OutputLayer->Weights[HiddenLayer->NumNrns * eON + eHN];

  File->Write( (void*)fBuf, ( (iBuf[0] * iBuf[1]) + (iBuf[1] * iBuf[2])) * 4);

  // Faz a limpeza
  delete[] fBuf;
  delete[] iBuf;
  delete File;
}

// -- ReadWeights -------------------------------------------------------------
bool LMBFeedforwardNetwork::ReadWeights (char* FileName)
{
  // Consulte Apndice 1 para conhecer o formato dos pesos gravados em disco
  TFileStream *File;
  unsigned short int *iBuf;
  float *fBuf;
  unsigned long int Pos = 0;

  File = new TFileStream (FileName, fmOpenRead);

  // L o cabealho:
  iBuf = new unsigned short int [3];
  File->Read((void*)iBuf, 6); // So 3 unsigned short int's -> 6 bytes

  // Verifica compatibilidade
  if ( iBuf[0] != InputLayer->NumNrns ||
       iBuf[1] != HiddenLayer->NumNrns ||
       iBuf[2] != OutputLayer->NumNrns )
      {
        ShowMessage("Arquivo Invlido - No  compatvel com esta rede"); // <=============== TIRAR DEPOIS DE OK!!!
        delete[] iBuf;
        return false;
      }

  // L os pesos:
  fBuf = new float [ (iBuf[0] * iBuf[1]) + (iBuf[1] * iBuf[2]) ];

  File->Read((void*)fBuf, ((iBuf[0] * iBuf[1]) + (iBuf[1] * iBuf[2])) * 4);

  for (int eHN = 0 ; eHN < iBuf[1] ; eHN++)
    for (int eIN = 0 ; eIN < iBuf[0] ; eIN++)
      HiddenLayer->Weights[iBuf[0] * eHN + eIN] = fBuf[Pos++];

  for (int eON = 0 ; eON < iBuf[2] ; eON++)
    for (int eHN = 0 ; eHN < iBuf[1] ; eHN++)
      OutputLayer->Weights[iBuf[1] * eON + eHN] = fBuf[Pos++];

  // Faz a limpeza
  delete[] fBuf;
  delete[] iBuf;
  delete File;

  return true;
}


//                  ------------------- X -------------------


/* ============================================================================
   =                          A P  N D I C E S                               =
   ============================================================================

 -> APNDICE 1: Formato dos pesos da rede gravados em arquivo:
      Os 6 primeiros bytes so 3 unsigned short int's, que representam, na
    ordem, o nmero de neurnios em cada uma das 3 camadas.
      Logo depois, uma sequncia de float's, na seguinte ordem: peso DO 1o.
    neurnio da 1a. camada PARA o 1o. neurnio da 2a. camada; peso DO 2o.
    neurnio da 1a. cadama PARA o 1o. neurnio da 2a. camada... at chegar ao
    peso DO n'simo neurnio da 1a. camada PARA o 1o. neurnio da 2a. camada.
    Neste ponto, todos o pesos PARA o 1o. neurnio da camada oculta esto
    prontos.
      A repete similarmente, em ordem, para todos os neurnios da 2a. camada.
      Terminados os pesos e bias da primeira para a segunda camada, vem os pesos
    e bias da segunda para a terceira, que seguem o mesmo esquema.
*/

#endif

// That's all Folks!

