/******************************************************************************
 *  LMBackprop.h - Classes para treinamento de rede neural feedforward com    *
 *                 o algortmo Backpropagation                                *
 *  Copyright(c) 1998 by LMBarros - Todos os direitos e esquerdos reservados  *
 *  Utiliza a classe TFileStream da VCL do C++ Builder...                     *
 ******************************************************************************/

#ifndef LMBackprop_h
#define LMBackprop_h

// Includes
#include <LMBFeedforward.h>

// ============================================================================
// = LMBackprop - Faz o treinamento propriamente dito                         =
// ============================================================================

// ----------------------------------------------------------------------------
// - Interface                                                                -
// ----------------------------------------------------------------------------

class LMBackprop
{
 public:
  LMBackprop(LMBFeedforwardNetwork *pNet);
  ~LMBackprop();
  float GetInput(int Pair, int Nrn);  // Se no houver nenhum par vai dar pau!...
  float GetOutput(int Pair, int Nrn); // ...Arrumar via HasAPair seria mais lento.
  void AddPair(float *Input, float *Output);
  float GetError(float *Targets);     // Retorna o erro mdio quadrtico (MSE)
  float GetEpochError();              // Retorna MSE para todos os pares
  void CalcGrads(int  Pair);          // Calcula gradientes para o par argumentado
  int Train(int MaxIter, float ErrTol, float LR, float *ErrVect); // Treina
  bool ReadData(char *FileName);      // retorna true se OK
  void WriteData(char *FileName);
  LMBFeedforwardNetwork *Net;         // Ponteiro para a rede a ser treinada
  float *Data;                        // Dados (entrada/saida desejada)
  float *GradsH, *GradsO;             // Gradientes da camada oculta / saida
  int NumPairs, InputSize, HiddenSize, OutputSize;
  bool HasAPair;
};

// ----------------------------------------------------------------------------
// - Implementao                                                            -
// ----------------------------------------------------------------------------

// - Construtor ---------------------------------------------------------------
LMBackprop::LMBackprop(LMBFeedforwardNetwork *pNet)
{
  NumPairs = 0;
  HasAPair = false;
  Net = pNet;
  InputSize = Net->InputLayer->NumNrns;
  HiddenSize = Net->HiddenLayer->NumNrns;
  OutputSize = Net->OutputLayer->NumNrns;
  GradsH = new float[HiddenSize * InputSize];
  GradsO = new float[OutputSize * HiddenSize];
}

// - Destrutor ----------------------------------------------------------------
LMBackprop::~LMBackprop()
{
  if (HasAPair) delete[] Data;
  delete[] GradsH;
  delete[] GradsO;
}

// - GetInput -----------------------------------------------------------------
float LMBackprop::GetInput(int Pair, int Nrn)
{
  return Data[(InputSize+OutputSize) * Pair + Nrn];
}

// - GetOutput ----------------------------------------------------------------
float LMBackprop::GetOutput(int Pair, int Nrn)
{
  return Data[(InputSize+OutputSize) * Pair + InputSize + Nrn];
}

// - AddPair ------------------------------------------------------------------
void LMBackprop::AddPair(float *Input, float *Output)
{
  float *vcTmp;
  int Pos = 0;
  int PairSize = InputSize + OutputSize;

  vcTmp = new float[PairSize * NumPairs];

  for (int c = 0 ; c < PairSize * NumPairs ; c++)  // Passa o Data
    vcTmp[c] = Data[c];                            // para o vcTmp

  if (HasAPair) delete[] Data;                     // Redimensiona o
  Data = new float[PairSize * (NumPairs + 1)];     // vetor Data

  for (int c = 0 ; c < PairSize * NumPairs ; c++)  // Copia de volta
    Data[c] = vcTmp[c];                            // no Data

  // Inclui o novo par, primeiro as entradas, depois as saidas
  for (int c = PairSize * NumPairs ; c < PairSize * NumPairs + InputSize; c++)
    Data[c] = Input[Pos++];

  Pos = 0;
  for (int c = PairSize * NumPairs + InputSize; c < PairSize * (NumPairs + 1) ; c++)
    Data[c] = Output[Pos++];

  NumPairs++; // Temos um par a mais
  HasAPair = true;

  delete[] vcTmp; // Limpa
}

// - GetError -----------------------------------------------------------------
float LMBackprop::GetError(float *Targets)
{
  float Err, Sum = 0;

  for (int c = 0 ; c < OutputSize ; c++)
  {
    Err = Targets[c] - Net->OutputLayer->Outputs[c];
    Sum += Err * Err;
  }
  return Sum / (float)OutputSize;
}

// - GetEpochError ------------------------------------------------------------
float LMBackprop::GetEpochError()
{
  float AccErr = 0, *Inp, *Trgt;

  Inp = new float[InputSize];
  Trgt = new float[OutputSize];

  for (int eP = 0 ; eP < NumPairs ; eP++)
  {
    for (int eI = 0 ; eI < InputSize ; eI++)   // Prepara vetor
      Inp[eI] = GetInput(eP, eI);              // entrada

    for (int eO = 0 ; eO < OutputSize ; eO++)  // Prepara vetor sada
      Trgt[eO] = GetOutput(eP, eO);            // desejada

    Net->GetOutputs(Inp);                      // Calcula Saidas

    AccErr += GetError(Trgt);                  // Acumula o erro mdio deste par
  }

  AccErr /= (float)NumPairs;

  delete[] Inp;
  delete[] Trgt;

  return AccErr;
}

// - CalcGrads ----------------------------------------------------------------
void LMBackprop::CalcGrads(int Pair)
{
  float Out, Delta, Sum, *Inp, *Trgt, *Deltas;

  Deltas = new float[OutputSize];
  Inp = new float[InputSize];
  Trgt = new float[OutputSize];

  // Prepara o par
  for (int eI = 0 ; eI < InputSize ; eI++)     // Prepara vetor
    Inp[eI] = GetInput(Pair, eI);              // entrada

  for (int eO = 0 ; eO < OutputSize ; eO++)    // Prepara vetor sada
    Trgt[eO] = GetOutput(Pair, eO);            // desejada

  Net->GetOutputs(Inp);

  // Camada de sada
  for (int eON = 0 ; eON < OutputSize ; eON++)
  {
    Out = Net->OutputLayer->Outputs[eON];
    Delta = (Trgt[eON] - Out) * ActDeriv(Out); // Eq. 6-11 pg 96
    Deltas[eON] = Delta;
    for (int eHN = 0 ; eHN < HiddenSize ; eHN++)  // Eq 6-12 pg 96
      GradsO[HiddenSize * eON + eHN] = Delta * Net->HiddenLayer->Outputs[eHN]; // para grd neg
  }

  // Camada oculta
  for (int eHN = 0 ; eHN < HiddenSize ; eHN++)
  {
    Sum = 0;
    for (int eON = 0 ; eON < OutputSize ; eON++)                             // Eq. 6-13
      Sum += Deltas[eON] * Net->OutputLayer->Weights[eON * HiddenSize + eHN];// pg 96

    Delta = Sum * ActDeriv(Net->HiddenLayer->Outputs[eHN]);                  // ainda a mesma

    for (int eIN = 0 ; eIN < InputSize ; eIN++)
      GradsH[InputSize * eHN + eIN] = Delta * Net->InputLayer->Outputs[eIN]; // Eq. 6-14 pg 96
  }

  delete[] Deltas;
  delete[] Inp;
  delete[] Trgt;
}

// - Train --------------------------------------------------------------------
int LMBackprop::Train (int MaxIter, float ErrTol, float LR, float *ErrVect)
{
  float Corr, Err;

  if (NumPairs == 0) return -1;

  for (int Iter = 0 ; Iter < MaxIter ; Iter++) // Uma vez p/ cada iterao
  {
    Err = GetEpochError();

    if (Err < ErrTol)
    {
      for (int c = Iter ; c < MaxIter ; c++)   // Preenche os espaos livres...
        ErrVect[c] = Err;                      // ... de ErrVect com AccErr
      return Iter + 1; // ... e cai fora.
    }

    // Seno tem que continuar o processo; s grava AccErr e segue
    ErrVect[Iter] = Err;

    for (int eP = 0 ; eP < NumPairs ; eP++)
    {
      // Depois, vai corrigindo os pesos:
      CalcGrads(eP);

      // Atualizao dos pesos, primeiro camada de saida...
      for (int eOW = 0 ; eOW <  OutputSize * HiddenSize ; eOW++)
      {
        Corr = LR * GradsO[eOW];
        Net->OutputLayer->Weights[eOW] += Corr;
      }

      // ...depois a escondida
      for (int eHW = 0 ; eHW < HiddenSize * InputSize ; eHW++)
      {
        Corr = LR * GradsH[eHW];
        Net->HiddenLayer->Weights[eHW] += Corr;
      }
    }
  }
  return MaxIter;
}

// - WriteData ----------------------------------------------------------------
void LMBackprop::WriteData(char *FileName)
{
  // Consulte Apndice 1 para ver o formato dos dados gravados em disco
  TFileStream *File;
  unsigned short int *iBuf;
  float *fBuf;

  File = new TFileStream (FileName, fmCreate);

  // Escreve o cabealho:
  iBuf = new unsigned short int [3];
  iBuf[0] = (unsigned short int)NumPairs;
  iBuf[1] = (unsigned short int)InputSize;
  iBuf[2] = (unsigned short int)OutputSize;
  File->Write((void*)iBuf, 6); // So 3 unsigned short int's -> 6 bytes

  // Escreve os dados:
  fBuf = new float [(InputSize + OutputSize) * NumPairs];

  for (unsigned short int ePair = 0 ; ePair < NumPairs ; ePair++ )
  {
    for (unsigned short int eInp = 0 ; eInp < InputSize ; eInp++ )
      fBuf[(ePair * (InputSize + OutputSize)) + eInp] = GetInput(ePair, eInp);
    for (unsigned short int eOut = 0 ; eOut < OutputSize ; eOut++ )
      fBuf[(ePair * (InputSize + OutputSize)) + InputSize + eOut] = GetOutput(ePair, eOut);
  }

  File->Write((void*)fBuf, (InputSize + OutputSize) * NumPairs * 4);

  // Faz a limpeza
  delete[] fBuf;
  delete[] iBuf;
  delete File;

}

// - ReadData ----------------------------------------------------------------
bool LMBackprop::ReadData (char *FileName)
{
  // Consulte Apndice 1 para conhecer o formato dos dados gravados em disco

  TFileStream *File;
  unsigned short int *iBuf;
  float *fBuf, *NewInput, *NewOutput;
  int PairSize, NewNumPairs;

  File = new TFileStream (FileName, fmOpenRead);

  // L o cabealho:
  iBuf = new unsigned short int[3];  // So 3 unsigned short int's...
  File->Read((void*)iBuf, 6);        // ... o mesmo que 6 bytes

  // Verificao de acordo com o cabealho:
  if (iBuf[1] != InputSize && iBuf[2] != OutputSize)
    return false;

  NewInput = new float[InputSize];
  NewOutput = new float[OutputSize];

  // Ajeita diversas variveis:
  NewNumPairs = iBuf[0];
  PairSize = InputSize + OutputSize;
//  if (HasAPair) delete[] Data;
//  Data = new float[PairSize * NewNumPairs];

  // L os dados:
  fBuf = new float [PairSize * NewNumPairs];

  File->Read((void*)fBuf, PairSize * NewNumPairs * 4);

  for (int ePair = 0 ; ePair < NewNumPairs ; ePair++ )
  {
    for (int eInp = 0 ; eInp < InputSize ; eInp++ )
      NewInput[eInp] = fBuf[ePair * PairSize + eInp];
    for (int eOut = 0 ; eOut < OutputSize ; eOut++ )
      NewOutput[eOut] = fBuf[ePair * PairSize + InputSize + eOut];
    AddPair(NewInput, NewOutput);
  }

  // Faz a limpeza
  delete[] fBuf;
  delete[] iBuf;
  delete[] NewInput;
  delete[] NewOutput;
  delete File;

  return true;
}


//                  ------------------- X -------------------


/* ============================================================================
   =                          A P  N D I C E S                               =
   ============================================================================

 -> APNDICE 1: Formato dos dados de treinamento gravados em arquivo
      Os 6 primeiros bytes so 3 unsigned short int's, que representam,
    respectivamente, o nmero de pares, o tamanho da entrada e o tamanho da
    sada desejada;
       A seguir, uma sequncia de float's, na seguinte ordem: dados de entrada
    do 1o. par; sada desejada do 1o. par; entrada do 2o. par; e assim por
    diante, numa ordem que parece ser lgica...

 -> APNDICE 2: Utilizao de LMBackprop::Train
      O prottipo da funo :
      void Train (int MaxIter, float ErrTol, float LR, float *ErrVect)
      MaxIter  o nmero mximo de iteraes. ErrTol  a tolerncia; se o erro
    cair a um valor menor que este, o treinamento pra. LR, learning rate,  a
    taxa de aprendizado. ErrVect  um ponteiro para um vetor de floats; nele
    iro os erros a cada iterao, para permitir uma anlise da convergncia; 
    importante ressaltar que este vetor deve ser previamente dimensionado para
    ter MaxIter elementos.


*/

#endif

// That's all Folks!

